import multiprocessing as mp
import numpy as np
import h5py as hp
import time
import sys

Nx, Ny, Nz = 0, 0, 0
u, v, w = 0, 0, 0

def readData():
    global Nx, Ny, Nz
    global u, v, w

    #fileName = "U.V1r.h5"
    #f = hp.File(fileName, 'r')
    #F_key = list(f.keys())[0]
    #u = np.array(f[F_key])
    #f.close()

    #fileName = "U.V2r.h5"
    #f = hp.File(fileName, 'r')
    #F_key = list(f.keys())[0]
    #v = np.array(f[F_key])
    #f.close()

    #fileName = "U.V3r.h5"
    #f = hp.File(fileName, 'r')
    #F_key = list(f.keys())[0]
    #w = np.array(f[F_key])
    #f.close()

    Nx, Ny, Nz = 64, 64, 64
    u = np.zeros((Nx, Ny, Nz))
    v = np.zeros((Nx, Ny, Nz))
    w = np.zeros((Nx, Ny, Nz))
    for i in range(64):
        for j in range(64):
            for k in range(64):
                u[i, j, k] = np.random.random()
                v[i, j, k] = np.random.random()
                w[i, j, k] = np.random.random()

    Nx, Ny, Nz = u.shape


def computeEigVals(xStr, xEnd):
    global u, v, w
    global Nx, Ny, Nz

    dx, dy, dz = 1.0/Nx, 1.0/Ny, 1.0/Nz

    xInd = 0
    subNx = xEnd - xStr

    c1 = np.zeros((subNx, Ny-2, Nz-2))
    c2 = np.zeros((subNx, Ny-2, Nz-2))
    c3 = np.zeros((subNx, Ny-2, Nz-2))

    for i in range(xStr, xEnd):
        for j in range(1, Ny-1):
            for k in range(1, Nz-1):
                dudx = (u[i+1, j, k] - u[i-1, j, k])/(2.0*dx)
                dudy = (u[i, j+1, k] - u[i, j-1, k])/(2.0*dy)
                dudz = (u[i, j, k+1] - u[i, j, k-1])/(2.0*dz)

                dvdx = (v[i+1, j, k] - v[i-1, j, k])/(2.0*dx)
                dvdy = (v[i, j+1, k] - v[i, j-1, k])/(2.0*dy)
                dvdz = (v[i, j, k+1] - v[i, j, k-1])/(2.0*dz)

                dwdx = (w[i+1, j, k] - w[i-1, j, k])/(2.0*dx)
                dwdy = (w[i, j+1, k] - w[i, j-1, k])/(2.0*dy)
                dwdz = (w[i, j, k+1] - w[i, j, k-1])/(2.0*dz)

                S = np.matrix([[dudx, (dudy + dvdx)*0.5, (dudz + dwdx)*0.5],
                               [(dvdx + dudy)*0.5, dvdy, (dvdz + dwdy)*0.5],
                               [(dwdx + dudz)*0.5, (dwdy + dvdz)*0.5, dwdz]])

                eVals, eVecs = np.linalg.eig(S)

                # Sort the eigenvalues in decreasing order
                sortedIndices = np.argsort(eVals)[::-1]

                c1[xInd, (j-1), (k-1)] = eVals[0]
                c2[xInd, (j-1), (k-1)] = eVals[1]
                c3[xInd, (j-1), (k-1)] = eVals[2]

        xInd += 1

    c1 = c1.flatten()
    c2 = c2.flatten()
    c3 = c3.flatten()

    return c1, c2, c3


def main():
    global Nx

    nProcs = 1
    try:
        nProcs = int(sys.argv[1])
        print("\nProcessing data using " + str(nProcs) + " processors\n")
    except:
        print("\nProcessing data in serial computation\n")

    readData()

    #print("Finished reading data\n")

    pool = mp.Pool(processes=nProcs)

    # Create range list along X direction
    rangeDivs = [int(x) for x in np.linspace(1, Nx-1, nProcs+1)]
    rangeList = [(rangeDivs[x], rangeDivs[x+1]) for x in range(nProcs)]

    #print("Indices of domains divisions: ", rangeDivs)
    #print("")
    #print("Range of indices for each sub-domain:", rangeList)
    #print("")

    # Parallel section
    strTime = time.time()
    poolRes = [pool.apply_async(computeEigVals, args=(x[0], x[1])) for x in rangeList]
    eigVals = [x.get() for x in poolRes]
    endTime = time.time()

    # Flatten the lists
    eig1 = [x for y in eigVals for x in y[0]]
    eig2 = [x for y in eigVals for x in y[1]]
    eig3 = [x for y in eigVals for x in y[2]]

    print("Finished computing eigen-values\n")
    print("Time taken to compute eigen-values = " + str(endTime - strTime))
    print("")

    # Generating PDF
    nBins = 100
    x = np.linspace(-1.0, 1.0, nBins)
    y1 = np.histogram(eig1, bins=nBins, range=(-1.0, 1.0), density=True)
    y2 = np.histogram(eig2, bins=nBins, range=(-1.0, 1.0), density=True)
    y3 = np.histogram(eig3, bins=nBins, range=(-1.0, 1.0), density=True)

    A = np.zeros((nBins, 4))
    A[:,0] = x
    A[:,1] = y1[0]
    A[:,2] = y2[0]
    A[:,3] = y3[0]
    np.savetxt('eigList.dat', A)


main()
